package org.processmining.traceclustering.algorithms.evolution.impl;

import java.util.ArrayList;
import java.util.Date;
import java.util.TreeMap;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.processmining.markovclustering.models.MCLMatrix;
import org.processmining.markovclustering.utils.MatrixUtils;
import org.processmining.traceclustering.algorithms.affinity.impl.CosineSimilarity;
import org.processmining.traceclustering.models.impl.AgeVector;
import org.processmining.traceclustering.models.impl.CaseVector;
import org.processmining.traceclustering.models.impl.TimeWindowType;
import org.processmining.traceclustering.models.perspectives.AbstractPerspective;
import org.processmining.traceclustering.parameters.impl.TimeWindowParameters;
import org.processmining.traceclustering.utils.LogUtils;

public class SimilarityEvolution {

	// For every time window, a new similarity matrix should be made and compared to the previous one
	// The results are stored in an object that is used to visualize in a histogram
	public static TimeSeriesCollection calculate(XLog eventlog, TimeWindowParameters timeWindowParameters,
			ArrayList<AbstractPerspective> perspectives) {

		System.out.println("Calculating change in behavior over time.");
		System.out.println("Time window type: " + timeWindowParameters.getType() + ", size: "
				+ timeWindowParameters.getWindowSize());

		TreeMap<Date, ArrayList<XEvent>> orderedEvents = LogUtils.getOrderedEvents(eventlog);

		final TimeSeries series = new TimeSeries("Change in behavior");

		int count = 0;
		int windows = 0;
		Date firstDate = null, lastDate = null;

		AgeVector ages = new AgeVector(eventlog.size(), 0.0);

		double change = 0.0;
		if (timeWindowParameters.getType().equals(TimeWindowType.NUMBEROFEVENTS)) {
			Date previousDate = null;
			MCLMatrix previousSimilarityMatrix = null;

			for (Date date : orderedEvents.keySet()) {
				if (firstDate == null)
					firstDate = date;

				for (int e = 0; e < orderedEvents.get(date).size(); e++) {
					if (count % timeWindowParameters.getWindowSize() == 0) {
						XLog filteredLog = LogUtils.filterLogUntil(eventlog, date, true);
						ArrayList<CaseVector> caseProfileVectors = CaseVector.mapLogToCaseVectors(filteredLog,
								perspectives);
						MCLMatrix similarityMatrix = new MCLMatrix(caseProfileVectors.size());

						CosineSimilarity cosSim = new CosineSimilarity();

						// Matrix is symmetrical, so only compute affinity for vectors above diagonal.
						for (int i = 0; i < caseProfileVectors.size(); i++) {
							for (int j = i; j < caseProfileVectors.size(); j++) {
								if (i != j) {
									double similarity = cosSim.computeSimilarity(caseProfileVectors.get(i),
											caseProfileVectors.get(j));
									similarityMatrix.set(i, j, similarity); // TOP TRIANGLE OF MATRIX
									similarityMatrix.set(j, i, similarity); // BOTTOM TRIANGLE OF MATRIX
								} else
									// The matrix is symmetrical, diagonal values are 1.
									similarityMatrix.set(i, i, 1d);
							}
						}

						// Set age factor to 1.0 for new traces
						int newCases = 0;
						for (int i = 0; i < filteredLog.size(); i++) {
							if (ages.get(i) == 0.0 && filteredLog.get(i).size() > 0) {
								ages.set(i, 1.0);
								newCases++;
							}
						}
						System.out.println("In this window there were " + newCases + " new cases.");

						if (count == 0) {
							// just set similarity matrix of first window, we don't have a window to compare change against yet
							System.out.println("Cannot compute change for first window.");
						} else {
							// Add value for every period after the first one

							// Multiply with age vector
							for (int index = 0; index < ages.size(); index++) {
								if (ages.get(index) == 0.0)
									assert (filteredLog.get(index).size() == 0);
								else if (ages.get(index) != 1.0) {
									for (int i = 0; i < previousSimilarityMatrix.numRows(); i++) {
										if (i != index) {
											previousSimilarityMatrix.set(i, index,
													previousSimilarityMatrix.get(i, index) * ages.get(index));
											previousSimilarityMatrix.set(index, i,
													previousSimilarityMatrix.get(index, i) * ages.get(index));
										}
									}
								}
							}

							// Then diminish ages for all traces
							ages.scale(0.95);

							change = MatrixUtils.getTotalPercentageChange(previousSimilarityMatrix, similarityMatrix);
							series.addOrUpdate(new org.jfree.data.time.Second(date), change);
							System.out.println("Change for window " + (windows) + ":\t" + previousDate + ",\t" + date
									+ ":\t" + change);
						}

						previousDate = (Date) date.clone();
						previousSimilarityMatrix = similarityMatrix.copy();

						windows++;
					}
					count++;
				}
				lastDate = date;
			}

			// Add value for the time from the last window to the end of the log (size is less than a full window)
			ArrayList<CaseVector> caseProfileVectors = CaseVector.mapLogToCaseVectors(eventlog, perspectives);
			MCLMatrix similarityMatrix = new MCLMatrix(caseProfileVectors.size());

			CosineSimilarity cosSim = new CosineSimilarity();

			// Matrix is symmetrical, so only compute affinity for vectors above diagonal.
			for (int i = 0; i < caseProfileVectors.size(); i++) {
				for (int j = i; j < caseProfileVectors.size(); j++) {
					if (i != j) {
						double similarity = cosSim.computeSimilarity(caseProfileVectors.get(i),
								caseProfileVectors.get(j));
						similarityMatrix.set(i, j, similarity); // TOP TRIANGLE OF MATRIX
						similarityMatrix.set(j, i, similarity); // BOTTOM TRIANGLE OF MATRIX
					} else
						// The matrix is symmetrical, diagonal values are 1.
						similarityMatrix.set(i, i, 1d);
				}
			}

			// Set age factor to 1.0 for new traces
			int newCases = 0;
			for (int i = 0; i < eventlog.size(); i++) {
				if (ages.get(i) == 0.0 && eventlog.get(i).size() > 0) {
					ages.set(i, 1.0);
					newCases++;
				}
			}
			System.out.println("In this window there were " + newCases + " new cases.");

			// Multiply with age vector
			for (int index = 0; index < ages.size(); index++) {
				if (ages.get(index) == 0.0)
					assert (eventlog.get(index).size() == 0);
				else if (ages.get(index) != 1.0) {
					for (int i = 0; i < previousSimilarityMatrix.numRows(); i++) {
						if (i != index) {
							previousSimilarityMatrix.set(i, index,
									previousSimilarityMatrix.get(i, index) * ages.get(index));
							previousSimilarityMatrix.set(index, i,
									previousSimilarityMatrix.get(index, i) * ages.get(index));
						}
					}
				}
			}

			// Then diminish ages for all traces
			ages.scale(0.95);

			change = MatrixUtils.getTotalPercentageChange(previousSimilarityMatrix, similarityMatrix);
			series.addOrUpdate(new org.jfree.data.time.Second(lastDate), change);
			System.out.println(
					"Change for window " + (windows) + ":\t" + previousDate + ",\t" + lastDate + ":\t" + change);

		} else if (timeWindowParameters.getType().equals(TimeWindowType.NUMBEROFMINUTES)) {
			// do something
		}

		System.out.println("Your log has " + count + " events.");
		System.out.println("Evolution has found " + windows + " windows with the given settings.");

		final TimeSeriesCollection timeWindowValues = new TimeSeriesCollection();
		timeWindowValues.addSeries(series);
		return timeWindowValues;
	}
}
