package org.processmining.traceclustering.models.impl;

import java.util.ArrayList;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.traceclustering.models.perspectives.AbstractPerspective;

import no.uib.cipr.matrix.sparse.SparseVector;

/**
 * Vector for keeping track of mapping from a trace to a certain perspective
 * 
 * @author B.F.A. Hompes
 *
 */
public class CaseVector extends SparseVector {

	private static final long serialVersionUID = -6420298104250532851L;

	public CaseVector(int size) {
		super(size);
	}

	private CaseVector(SparseVector vector) {
		super(vector);
	}

	@Override
	public CaseVector copy() {
		return new CaseVector(super.copy());
	}

	public CaseVector toUnitVector() {
		if (isZeroVector())
			return this;

		CaseVector unit = copy();
		double magnitude = unit.norm1();
		for (int i = 0; i < unit.size; i++) {
			unit.set(i, unit.get(i) / magnitude);
		}
		return unit;
	}

	public boolean isZeroVector() {
		for (double element : this.getData())
			if (element != 0.0)
				return false;
		return true;
	}

	public static CaseVector concat(ArrayList<CaseVector> vectors) {
		int size = 0;

		for (CaseVector vector : vectors)
			size += vector.size();

		CaseVector combination = new CaseVector(size);

		int index = 0;
		for (CaseVector vector : vectors) {
			for (int i = 0; i < vector.size; i++)
				combination.set(index + i, vector.get(i));
			index += vector.size();
		}

		return combination;
	}

	/**
	 * Two casevectors are equal if they have the same length and elements at
	 * corresponding positions.
	 * 
	 * (Is actually already handled by the base class SparseVector.)
	 */
	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof CaseVector))
			return false;

		CaseVector vect = (CaseVector) obj;

		if (size() != vect.size())
			return false;

		for (int i = 0; i < this.size; i++) {
			if (get(i) != vect.get(i))
				return false;
		}

		return true;
	}

	public static ArrayList<CaseVector> mapLogToCaseVectors(XLog eventlog,
			ArrayList<AbstractPerspective> perspectives) {
		ArrayList<CaseVector> caseProfileVectors = new ArrayList<CaseVector>();

		// Walk through the log one time
		for (XTrace trace : eventlog) {
			ArrayList<CaseVector> vectors = new ArrayList<CaseVector>();
			// For every perspective in the set of selected perspectives, map the trace
			for (AbstractPerspective perspective : perspectives) {
				if (perspective.isSelected()) {
					vectors.add(perspective.projectCase(eventlog, trace));
				}
			}
			CaseVector vector = CaseVector.concat(vectors);
			caseProfileVectors.add(vector);
		}

		return caseProfileVectors;
	}
}