package org.processmining.traceclustering.plugins;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

import org.deckfour.xes.classification.XEventAttributeClassifier;
import org.deckfour.xes.info.XAttributeInfo;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.basicutils.parameters.PluginParameters;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.markovclustering.algorithms.clustering.impl.MCLAlgorithm_JAMA_Par;
import org.processmining.markovclustering.models.MCLMatrix;
import org.processmining.markovclustering.utils.MatrixUtils;
import org.processmining.traceclustering.algorithms.affinity.impl.CosineSimilarity;
import org.processmining.traceclustering.connections.CaseClusteringConnection;
import org.processmining.traceclustering.models.impl.CaseCluster;
import org.processmining.traceclustering.models.impl.CaseClustering;
import org.processmining.traceclustering.models.impl.CaseVector;
import org.processmining.traceclustering.models.perspectives.AbstractPerspective;
import org.processmining.traceclustering.models.perspectives.impl.EventAlphabetFrequencyPerspective;
import org.processmining.traceclustering.models.perspectives.impl.EventAlphabetOccurrencePerspective;
import org.processmining.traceclustering.models.perspectives.impl.TraceAlphabetOccurrencePerspective;
import org.processmining.traceclustering.parameters.CaseClusteringParameters;
import org.processmining.traceclustering.utils.StringUtils;

/**
 * Plug-in for the clustering of cases in an event log based on a set of
 * selected perspectives, using the Markov clustering algorithm.
 * 
 * @author B.F.A. Hompes
 * 
 */
@Plugin(
		name = "Cluster cases using Markov clustering",
		parameterLabels = { "Event log", "MCL clustering parameters" },
		returnLabels = { "Case clustering" },
		returnTypes = { CaseClustering.class },
		categories = { PluginCategory.Discovery, PluginCategory.Analytics, PluginCategory.Filtering },
		keywords = "MCL, markov clustering, case clustering, trace clustering"
				+ "sequence clustering, outlier detection, mainstream behavior, deviating behavior",
		help = "Find groups of cases that are similar on a selected set of perspectives "
				+ "using the MCL Markov Clustering algorithm. "
				+ "The algorithm does not take the amount of clusters as input, "
				+ "hence it is able to find both mainstream and deviating behavior.",
		userAccessible = true,
		handlesCancel = false,
		mostSignificantResult = 1)
public class CaseClusteringPlugin {

	/***************************************************************
	 * 
	 * Clusters the cases in the event log with Markov clustering, using default
	 * parameters.
	 * 
	 * @param context
	 *            The UI plug-in context
	 * @param eventlog
	 *            The event log to cluster
	 * @return The case clustering (visualized)
	 * 
	 **************************************************************/
	@UITopiaVariant(
			affiliation = UITopiaVariant.EHV,
			author = "B.F.A. Hompes",
			email = "b.f.a.hompes@tue.nl")
	@PluginVariant(
			variantLabel = "Cluster cases using Markov clustering, simple",
			requiredParameterLabels = { 0 },
			help = "Cluster cases using Markov clustering, with default parameters.")
	public CaseClustering doClusteringPublic(PluginContext context, XLog eventlog) {

		CaseClusteringParameters parameters = new CaseClusteringParameters();
		return doClusteringPrivate(context, eventlog, parameters);
	}

	/**************************************************************
	 * 
	 * Clusters the cases in the event log with Markov clustering, using the
	 * given parameters.
	 * 
	 * @param context
	 *            The UI plug-in context
	 * @param eventlog
	 *            The event log to cluster
	 * @param parameters
	 *            The parameters for the case clustering algorithm
	 * @return The case clustering (visualized)
	 * 
	 **************************************************************/
	@UITopiaVariant(
			affiliation = UITopiaVariant.EHV,
			author = "B.F.A. Hompes",
			email = "b.f.a.hompes@tue.nl")
	@PluginVariant(
			variantLabel = "Cluster cases using Markov clustering, parameters",
			requiredParameterLabels = { 0, 1 },
			help = "Cluster cases using Markov clustering, with manually set parameters.")
	public CaseClustering doClusteringPublic(PluginContext context, XLog eventlog,
			CaseClusteringParameters parameters) {

		return doClusteringPrivate(context, eventlog, parameters);
	}

	/**************************************************************
	 * 
	 * Clusters the cases in the event log with Markov clustering, using the
	 * given parameters.
	 * 
	 * @param context
	 *            The UI plug-in context
	 * @param eventlog
	 *            The event log to cluster
	 * @param parameters
	 *            The parameters for the case clustering algorithm
	 * @return The case clustering (visualized)
	 * 
	 **************************************************************/
	private CaseClustering doClusteringPrivate(PluginContext context, XLog eventlog,
			CaseClusteringParameters parameters) {

		// We do not actively keep track of progress
		context.getProgress().setIndeterminate(true);

		// Start counting running time
		System.out.println("Starting MCL Clustering plugin.");
		context.log("Starting MCL Clustering plugin.");
		long start = System.currentTimeMillis();

		/**************************************************************
		 *
		 * Check whether or not there's a connection available so we don't have
		 * to run everything again.
		 *
		 **************************************************************/

		if (parameters.isTryConnections() && parameters.getPerspectives().size() != 0) {
			context.log("Checking for connections.");
			Collection<CaseClusteringConnection> connections;
			try {
				connections = context.getConnectionManager().getConnections(CaseClusteringConnection.class, context,
						eventlog);
				for (CaseClusteringConnection connection : connections) {
					if (connection.getObjectWithRole(CaseClusteringConnection.LOG).equals(eventlog)) {
						CaseClustering result = (CaseClustering) connection
								.getObjectWithRole(CaseClusteringConnection.CASECLUSTERING);
						if (result.getCaseClusteringParameters().equals(parameters)) {
							System.out.println("Connection found. Returning the stored clustering.");
							context.log("Connection found. Returning the stored clustering.");
							return connection.getObjectWithRole(CaseClusteringConnection.CASECLUSTERING);
						}
					}
				}
			} catch (ConnectionCannotBeObtained e) {
				System.out.println("No connections found. Start clustering.");
				context.log("No connections found. Start clustering.");
			}
		}

		/**************************************************************
		 *
		 * Create clustering
		 *
		 **************************************************************/

		CaseClustering clustering = new CaseClustering();
		clustering.setEventlog(eventlog);
		clustering.setCaseClusteringParameters(parameters);

		/**************************************************************
		 *
		 * Make a selection of attributes / perspectives / profile to cluster
		 * upon
		 *
		 **************************************************************/

		XLogInfo logInfo = XLogInfoFactory.createLogInfo(eventlog);

		// If there are no perspectives in the clustering parameters object yet we need to find the perspectives first
		// If there are already perspectives in the parameters object then we've already done this.
		if (parameters.getPerspectives().size() == 0) {
			// Keep track of case attribute info
			XAttributeInfo caseAttributeInfo = logInfo.getTraceAttributeInfo();
			HashMap<String, HashSet<XAttribute>> caseAttributeKeyValues = new HashMap<String, HashSet<XAttribute>>();

			// Loop through the log once to find all values for a given case attribute
			for (XTrace trace : eventlog) {
				for (String attributeKey : caseAttributeInfo.getAttributeKeys()) {
					if (caseAttributeKeyValues.get(attributeKey) == null)
						caseAttributeKeyValues.put(attributeKey, new HashSet<XAttribute>());
					if (trace.getAttributes().get(attributeKey) != null)
						caseAttributeKeyValues.get(attributeKey).add(trace.getAttributes().get(attributeKey));
				}
			}

			// Create a case attribute perspective for all case attributes
			for (String attributeKey : caseAttributeInfo.getAttributeKeys()) {
				AbstractPerspective perspective = new TraceAlphabetOccurrencePerspective("Trace attr.: " + attributeKey,
						logInfo, attributeKey, caseAttributeKeyValues.get(attributeKey));
				parameters.getPerspectives().add(perspective);
			}

			//TODO [medium] check below 
			// Create a new XEventAttributeClassifier for each attribute in the log
			// These will be used as alphabet perspectives
			XAttributeInfo eventAttributeInfo = logInfo.getEventAttributeInfo();
			HashSet<String> eventAttributes = new HashSet<String>(eventAttributeInfo.getAttributeKeys());
			for (String attribute : eventAttributes) {
				// Create a classifier for this event attribute
				XEventAttributeClassifier attributeClassifier = new XEventAttributeClassifier(attribute + "Classifier",
						attribute);
				// Create perspective from classifier and add to set
				logInfo = XLogInfoFactory.createLogInfo(eventlog, attributeClassifier);
				AbstractPerspective perspective = new EventAlphabetFrequencyPerspective(
						"Event attr. frequency: " + attribute, logInfo, attributeClassifier);
				AbstractPerspective perspective2 = new EventAlphabetOccurrencePerspective(
						"Event attr. occurrence: " + attribute, logInfo, attributeClassifier);
				parameters.getPerspectives().add(perspective);
				parameters.getPerspectives().add(perspective2);
			}

			// Here we can add any other perspective imaginable, potentially case attributes and/or computed perspectives etc.

			// We can now return an empty clustering.
			// This will let the user decide which perspectives to cluster upon.			
			return clustering;
		}

		System.out.println("Expansion: " + parameters.getMclParameters().getExpansionParameter());
		System.out.println("Inflation: " + parameters.getMclParameters().getInflationParameter());

		/**************************************************************
		 *
		 * Map all cases to the chosen perspectives
		 *
		 **************************************************************/
		System.out.println("Start mapping " + eventlog.size() + " cases to the chosen perspectives.");
		context.log("Start mapping " + eventlog.size() + " cases to the chosen perspectives.");

		ArrayList<CaseVector> caseProfileVectors = new ArrayList<CaseVector>();
		// Walk through the log one time
		for (XTrace trace : eventlog) {
			ArrayList<CaseVector> vectors = new ArrayList<CaseVector>();
			// For every perspective in the set of selected perspectives, map the trace
			for (AbstractPerspective perspective : parameters.getPerspectives()) {
				if (perspective.isSelected()) {
					vectors.add(perspective.projectCase(eventlog, trace));
				}
			}
			CaseVector vector = CaseVector.concat(vectors);
			caseProfileVectors.add(vector);
		}
		System.out.println("Cases mapped to the chosen perspectives.");
		context.log("Cases mapped to the chosen perspectives.");

		/**************************************************************
		 *
		 * Compute cosine similarity between cases
		 *
		 **************************************************************/

		System.out.println("Start computing case similarities.");
		context.log("Start computing case similarities.");

		// Create affinity / similarity matrix A of cases using cosine similarity
		MCLMatrix similarityMatrix = new MCLMatrix(caseProfileVectors.size());

		CosineSimilarity cosSim = new CosineSimilarity();

		// Matrix is symmetrical, so only compute affinity for vectors above diagonal.
		for (int i = 0; i < caseProfileVectors.size(); i++) {
			for (int j = i; j < caseProfileVectors.size(); j++) {
				if (i != j) {
					double similarity = cosSim.computeSimilarity(caseProfileVectors.get(i), caseProfileVectors.get(j));
					similarityMatrix.set(i, j, similarity); // TOP TRIANGLE OF MATRIX
					similarityMatrix.set(j, i, similarity); // BOTTOM TRIANGLE OF MATRIX
				} else
					// The matrix is symmetrical, diagonal values are 1.
					similarityMatrix.set(i, i, 1d);
			}
		}

		System.out.println("Similarity matrix calculated.");
		context.log("Similarity matrix calculated.");

		if (parameters.getMessageLevel() == PluginParameters.DEBUG)
			System.out.println(similarityMatrix.toString());

		/**************************************************************
		 *
		 * Create clustering
		 *
		 **************************************************************/

		// Set clustering
		clustering.setEventlog(eventlog);
		clustering.setCaseClusteringParameters(parameters);
		clustering.setSimilarityMatrix(similarityMatrix);

		// Optionally add self-loops (add self-loop parameter value on diagonal)
		//		affinityMatrix.addSelfLoop(1.0/eventlog.size());

		// Column-normalize the matrix
		MCLMatrix normalizedMatrix = similarityMatrix.copy();
		MatrixUtils.columnNormalize(normalizedMatrix);
		System.out.println("Matrix normalized.");
		//		System.out.println(normalizedMatrix.toString());
		context.log("Matrix normalized.");

		// Apply MCL clustering until the matrix converges (within bounds of parameters)
		context.log("Start MCL algorithm.");
		MCLMatrix clusteringMatrix = (new MCLAlgorithm_JAMA_Par()).apply(normalizedMatrix,
				parameters.getMclParameters());
		context.log("Finished MCL algorithm.");

		normalizedMatrix = null;

		// Interpret as clustering (find attractors, group with their attractees)
		// Attractors are vertices that have at least a positive value in their corresponding row	
		context.log("Interpreting clustering.");
		for (int row = 0; row < clusteringMatrix.numRows(); row++) {
			CaseCluster cluster = new CaseCluster();
			cluster.add(row);
			for (int column = 0; column < clusteringMatrix.numColumns(); column++) {
				if (clusteringMatrix.get(row, column) > parameters.getMclParameters().getStopCriterion()) {
					cluster.add(column);
				}
			}
			clustering.add(cluster);
		}
		clusteringMatrix = null;

		/**************************************************************
		 *
		 * Create annotations for clustering
		 *
		 **************************************************************/

		// Create annotations for clustering
		context.log("Creating annotations.");
		int totalNumberOfCases = 0;

		for (CaseCluster cluster : clustering) {
			totalNumberOfCases += cluster.size();

			for (AbstractPerspective perspective : parameters.getPerspectives()) {
				if (!perspective.isSelected())
					continue;

				cluster.getAnnotation().put(perspective, perspective.createAnnotation(eventlog, cluster));
			}
		}

		// End time calculation
		long end = System.currentTimeMillis();

		/**************************************************************
		 *
		 * Output
		 *
		 **************************************************************/

		// Print clustering
		if (parameters.getMessageLevel() == PluginParameters.DEBUG)
			clustering.print();

		// Do some logging
		System.out.println(clustering.size() + " clusters found.");
		System.out.println("Total operation took: " + StringUtils.FormatMillisToDate(end - start));
		System.out.println("Sum of cluster sizes: " + totalNumberOfCases);

		context.log("Total operation took: " + StringUtils.FormatMillisToDate(end - start));
		context.log(clustering.size() + " clusters found.");

		// Save clustering connection
		context.log("Saving clustering connection.");
		if (parameters.isTryConnections()) {
			context.getConnectionManager().addConnection(new CaseClusteringConnection(eventlog, clustering));
		}

		// Return clusters
		context.log("Finished CaseClustering plugin.");
		context.log("Showing resulting clustering.");

		return clustering;
	}
}