package org.processmining.traceclustering.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.TreeMap;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;

public class LogUtils {

	/**
	 * Return a sorted map of <Date, XEvent> for the given log.
	 * 
	 * @param eventlog
	 *            The event log
	 * @return The sorted map of events
	 */
	public static TreeMap<Date, ArrayList<XEvent>> getOrderedEvents(XLog eventlog) {
		TreeMap<Date, ArrayList<XEvent>> orderedEvents = new TreeMap<Date, ArrayList<XEvent>>();

		for (XTrace trace : eventlog) {
			for (XEvent event : trace) {
				Date eventDate = XTimeExtension.instance().extractTimestamp(event);
				if (eventDate != null) {
					if (!orderedEvents.keySet().contains(eventDate)) {
						orderedEvents.put(eventDate, new ArrayList<XEvent>());
					}
					orderedEvents.get(eventDate).add(event);
				}
			}
		}

		return orderedEvents;
	}

	/**
	 * Filter the log to only keep traces of which the indices are given.
	 * 
	 * @param eventlog
	 *            The original event log
	 * @param traceIndices
	 *            The indices of traces to keep
	 * @return The filtered event log
	 */
	public static XLog filterLog(XLog eventlog, Collection<Integer> traceIndices) {
		XLog filteredLog = (XLog) eventlog.clone();

		HashSet<XTrace> tracesToKeep = new HashSet<XTrace>();
		for (Integer index : traceIndices) {
			tracesToKeep.add(eventlog.get(index));
		}
		filteredLog.clear();
		filteredLog.addAll(tracesToKeep);

		return filteredLog;
	}

	/**
	 * Filter the log to only keep events which occurred before or at
	 * (inclusive=true) the indicated time. Traces can be cut to remove events
	 * occurring after the given time, or can be removed completely (when they
	 * have no events before the given time).
	 * 
	 * @param eventlog
	 *            The original event log
	 * @param timestamp
	 *            The time to look at
	 * @param inclusive
	 *            Whether or not to keep events that occurred exactly on the
	 *            given time.
	 * @return The filtered event log
	 */
	public static XLog filterLogUntil(XLog eventlog, Date timestamp, boolean inclusive) {
		XLog filteredLog = (XLog) eventlog.clone();
		filteredLog.clear();
		for (XTrace trace : eventlog) {
			// Start with an empty trace rather than a copy (because of trace attributes).
			XTrace traceCopy = XFactoryRegistry.instance().currentDefault().createTrace();
			traceCopy.clear();
			for (XEvent event : trace) {
				Date eventDate = XTimeExtension.instance().extractTimestamp(event);
				if (eventDate.compareTo(timestamp) < 0)
					traceCopy.add(event);
				if (inclusive && eventDate.compareTo(timestamp) == 0)
					traceCopy.add(event);
			}
			// If there's an event in the trace, also the trace attributes should be present.
			// If the trace is empty, no trace attributes should be available because that messes up similarity calculation.
			if (traceCopy.size() > 0)
				traceCopy.setAttributes((XAttributeMap) trace.getAttributes().clone());
			//			else
			//				System.out.println("didn't add empty trace");
			filteredLog.add(traceCopy);
		}
		return filteredLog;
	}

	/**
	 * Filter the log to only keep traces which were active at the given time.
	 * Traces can be cut to remove events occurring after the given time.
	 * 
	 * @param eventlog
	 *            The event log to filter.
	 * @param timestamp
	 *            The time to look at.
	 * @return The filtered event log (active traces)
	 */
	public static XLog filterLogActiveTraces(XLog eventlog, Date timestamp) {
		XLog filteredLog = (XLog) eventlog.clone();
		filteredLog.clear();
		for (XTrace trace : eventlog) {
			if (isActive(trace, timestamp))
				filteredLog.add(trace);
		}
		return filteredLog;
	}

	/**
	 * Checks whether the given trace is active at the given time. A trace is
	 * considered active at a given time when it has events both before and
	 * after that time or exactly on that time.
	 * 
	 * @param trace
	 *            The trace to check.
	 * @param timestamp
	 *            The time to look at.
	 * @return Whether or not the trace is active.
	 */
	public static boolean isActive(XTrace trace, Date timestamp) {
		boolean before = false, after = false;
		for (XEvent event : trace) {
			if (XTimeExtension.instance().extractTimestamp(event).before(timestamp))
				before = true;

			if (XTimeExtension.instance().extractTimestamp(event).after(timestamp))
				after = true;

			if (before && after)
				return true;

			if (XTimeExtension.instance().extractTimestamp(event).compareTo(timestamp) == 0)
				return true;
		}
		return false;
	}

}