package org.processmining.traceclustering.visualization;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executor;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.ListSelectionModel;
import javax.swing.border.EmptyBorder;

import org.apache.commons.collections15.functors.ConstantTransformer;
import org.apache.commons.collections15.functors.MapTransformer;
import org.apache.commons.collections15.map.LazyMap;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.traceclustering.connections.CaseClusteringConnection;
import org.processmining.traceclustering.models.impl.CaseCluster;
import org.processmining.traceclustering.models.impl.CaseClustering;
import org.processmining.traceclustering.models.perspectives.AbstractPerspective;
import org.processmining.traceclustering.parameters.CaseClusteringParameters;
import org.processmining.traceclustering.plugins.CaseClusteringPlugin;
import org.processmining.traceclustering.visualization.transformers.EdgeDrawPaintTransformer;
import org.processmining.traceclustering.visualization.transformers.VertexLabelTransformer;
import org.processmining.traceclustering.visualization.transformers.VertexShapeTransformer;
import org.processmining.traceclustering.visualization.transformers.VertexToolTipTransformer;

import com.fluxicon.slickerbox.factory.SlickerDecorator;
import com.fluxicon.slickerbox.factory.SlickerFactory;

import edu.uci.ics.jung.algorithms.layout.AggregateLayout;
import edu.uci.ics.jung.algorithms.layout.KKLayout;
import edu.uci.ics.jung.algorithms.layout.Layout;
import edu.uci.ics.jung.graph.util.Pair;
import edu.uci.ics.jung.visualization.DefaultVisualizationModel;
import edu.uci.ics.jung.visualization.VisualizationModel;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.DefaultModalGraphMouse;
import edu.uci.ics.jung.visualization.picking.PickedState;

/**
 * Visualizer for trace clustering.
 * 
 * @author B.F.A. Hompes
 *
 */
@Plugin(
		name = "Visualize Trace Clustering",
		returnLabels = { "Visualized Trace Clustering" },
		returnTypes = { JComponent.class },
		parameterLabels = { "Trace Clustering" },
		userAccessible = true)
@Visualizer
public class CaseClusteringVisualization {

	UIPluginContext globalContext;
	CaseClustering globalClustering;
	@SuppressWarnings("rawtypes")
	VisualizationViewer visView;
	DefaultModalGraphMouse<String, Object> graphMouse;
	SelectionDetailsPane rightPane;
	PickedState<CaseCluster> pickedState;

	// left and centerRight
	JSplitPane entireView = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
	// center and Right
	JSplitPane centerRight = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);

	private SlickerFactory slickerFactory;
	private SlickerDecorator slickerDecorator;

	@SuppressWarnings({ "unchecked", "rawtypes" })
	private final Map vertexPaints = LazyMap.decorate(new HashMap(), new ConstantTransformer(Color.gray));

	@SuppressWarnings({ "unchecked", "rawtypes" })
	@PluginVariant(
			requiredParameterLabels = { 0 })
	public JPanel visualize(UIPluginContext context, CaseClustering clustering) {
		/**************************************************************
		 * 
		 * Some global stuff
		 *
		 *************************************************************/

		globalContext = context;
		globalClustering = clustering;

		/**************************************************************
		 * 
		 * Default look & feel stuff & layout
		 *
		 *************************************************************/

		slickerFactory = SlickerFactory.instance();
		slickerDecorator = SlickerDecorator.instance();

		// Return graph in panel
		JPanel panel = new JPanel(new BorderLayout());
		entireView.setOpaque(false);
		centerRight.setOpaque(false);
		entireView.setOneTouchExpandable(true);
		centerRight.setOneTouchExpandable(true);

		/**************************************************************
		 *
		 * Left panel: dimension selection and calculate button
		 *
		 **************************************************************/

		JPanel leftPane = new JPanel();
		leftPane.setPreferredSize(new Dimension(250, 80));
		leftPane.setBackground(Color.DARK_GRAY);
		leftPane.setForeground(Color.WHITE);
		leftPane.setLayout(new BoxLayout(leftPane, BoxLayout.PAGE_AXIS));
		leftPane.setBorder(new EmptyBorder(15, 5, 5, 15));

		// Mouse information
		JLabel mouseInfo = slickerFactory.createLabel("Select the mouse mode.");
		mouseInfo.setForeground(leftPane.getForeground());
		leftPane.add(mouseInfo);

		// Default mouse support
		graphMouse = new DefaultModalGraphMouse<String, Object>();
		JComboBox mouseMode = graphMouse.getModeComboBox();
		slickerDecorator.decorate(mouseMode);
		leftPane.add(mouseMode);

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));

		// MCL Parameter inputs	
		CaseClusteringParameters parameters = globalClustering.getCaseClusteringParameters();
		JLabel expInfo = slickerFactory.createLabel("Expansion parameter.");
		expInfo.setForeground(leftPane.getForeground());
		leftPane.add(expInfo);
		final ProMTextField expansion = new ProMTextField(parameters.getMclParameters().getExpansionParameter() + "",
				"Expansion parameter (integer).");
		leftPane.add(expansion);

		JLabel inflInfo = slickerFactory.createLabel("Inflation parameter.");
		inflInfo.setForeground(leftPane.getForeground());
		leftPane.add(inflInfo);
		final ProMTextField inflation = new ProMTextField(parameters.getMclParameters().getInflationParameter() + "",
				"Inflation parameter (double).");
		leftPane.add(inflation);

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));

		// Dimensions information
		JLabel dimensionInfo = slickerFactory.createLabel("Select the clustering dimensions.");
		dimensionInfo.setForeground(leftPane.getForeground());
		leftPane.add(dimensionInfo);

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));

		// Dimensions selection
		Collections.sort(globalClustering.getCaseClusteringParameters().getPerspectives(),
				new Comparator<AbstractPerspective>() {
					@Override
					public int compare(AbstractPerspective p1, AbstractPerspective p2) {
						return p1.getDescription().compareToIgnoreCase(p2.getDescription());
					}
				});
		JList dimensionsBox = new JList(globalClustering.getCaseClusteringParameters().getPerspectives().toArray());
		dimensionsBox.setCellRenderer(new PerspectiveCellRenderer());
		dimensionsBox.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		dimensionsBox.setBackground(leftPane.getBackground());
		dimensionsBox.setForeground(leftPane.getForeground());
		JScrollPane dimensionsScroller = new JScrollPane(dimensionsBox);
		dimensionsScroller.setBorder(null);
		slickerDecorator.decorate(dimensionsScroller, leftPane.getBackground(), Color.LIGHT_GRAY, Color.LIGHT_GRAY);
		leftPane.add(dimensionsScroller);

		final JButton redoButton = slickerFactory.createButton("Calculate clustering");
		redoButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				final CaseClusteringParameters parameters = new CaseClusteringParameters(
						globalClustering.getCaseClusteringParameters());
				try {
					int expansionParameter = 0;
					double inflationParameter = 0;

					expansionParameter = Integer.parseInt(expansion.getText());
					inflationParameter = Double.parseDouble(inflation.getText());

					parameters.getMclParameters().setExpansionParameter(expansionParameter);
					parameters.getMclParameters().setInflationParameter(inflationParameter);
				} catch (Exception ex) {
					ex.printStackTrace();
				}

				final UIPluginContext child = globalContext.createChildContext("child");
				Executor exec = child.getExecutor();

				Runnable task = new Runnable() {
					public void run() {
						// Change button name to show it's doing something
						redoButton.setText("Calculating clustering...");

						CaseClusteringPlugin plugin = new CaseClusteringPlugin();
						CaseClustering clustering = plugin.doClusteringPublic(child, globalClustering.getEventlog(),
								parameters);

						// View the resulting clustering in the center pane of the window
						centerRight.setLeftComponent(createVisualisationViewer(clustering));

						// Save this new clustering as a provided object in the resource view (not as favorite)
						child.getProvidedObjectManager().createProvidedObject("Clustering", clustering,
								CaseClustering.class, globalContext);

						// Save connection between the log, the parameters and the resulting clustering
						child.getConnectionManager().addConnection(
								new CaseClusteringConnection(globalClustering.getEventlog(), clustering));

						centerRight.setRightComponent(createRightPanel());

						redoButton.setText("Calculate clustering");
					}
				};

				exec.execute(task);
			}
		});

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));
		leftPane.add(redoButton);

		/**************************************************************
		 *
		 * Center panel: clustering visualization (graph)
		 *
		 **************************************************************/

		visView = createVisualisationViewer(globalClustering);

		/**************************************************************
		 *
		 * Right panel: details / selection
		 *
		 **************************************************************/

		rightPane = createRightPanel();

		/**************************************************************
		 *
		 * Wrap everything together and return
		 *
		 **************************************************************/

		centerRight.setLeftComponent(visView);
		centerRight.setRightComponent(rightPane);

		entireView.setLeftComponent(leftPane);
		entireView.setRightComponent(centerRight);

		panel.add(entireView, BorderLayout.CENTER);

		return panel;
	}

	/**
	 *
	 * Will create a graph to visualize the clustering using the JUNG library.
	 *
	 * @author B.F.A. Hompes
	 * @param clustering
	 * 
	 * @return A visualization viewer for the clustering
	 *
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private VisualizationViewer createVisualisationViewer(CaseClustering clustering) {

		// Create the master graph in the JUNG library
		CaseClusterGraph graph = new CaseClusterGraph();

		// Add vertices (clusters)
		for (CaseCluster cluster : clustering) {
			graph.addVertex(cluster);
			// By default clusters are light gray
			vertexPaints.put(cluster, new Color(240, 240, 240));
		}

		// Add edges 
		// Similar traces are connected, only one edge is inserted for connections between clusters (traces in clusters)
		for (CaseCluster c1 : clustering) {
			cluster2: for (CaseCluster c2 : clustering) {
				if (!c1.equals(c2)) {
					// Only check one way since matrix is symmetrical over diagonal, hence matrix(s,t)=matrix(t,s)
					for (Integer source : c1) {
						for (Integer target : c2) {
							if (clustering.getSimilarityMatrix().get(source, target) > 1E-4) {
								graph.addEdge(new Edge(c1, c2, clustering.getSimilarityMatrix().get(source, target)),
										new Pair(c1, c2));
								break cluster2;
							}
						}
					}
				}
			}
		}

		// Create a general layout for our master graph, using the specified layout algorithm
		Layout layout = new KKLayout(graph);

		// Create an aggregate layout because we need subgraphs (for the clusters)
		AggregateLayout aglayout = new AggregateLayout(layout);

		// Visualization of the layout
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension preferredSize = new Dimension((int) (screenSize.width * 0.9), (int) (screenSize.height * 0.85));
		VisualizationModel visualizationModel = new DefaultVisualizationModel(aglayout, preferredSize);
		visView = new VisualizationViewer(visualizationModel, preferredSize);

		// Tell the renderer to use our own customized color rendering
		visView.getRenderContext().setVertexFillPaintTransformer(MapTransformer.getInstance(vertexPaints));

		// Transformer for the vertices (label)
		visView.getRenderContext().setVertexLabelTransformer(new VertexLabelTransformer());

		// Transformer for the vertices (shape)
		visView.getRenderContext().setVertexShapeTransformer(new VertexShapeTransformer());

		// Transformer for the vertices (tooltip)
		visView.setVertexToolTipTransformer(new VertexToolTipTransformer());

		//  Transformer for the edges (color)
		visView.getRenderContext()
				.setEdgeDrawPaintTransformer(new EdgeDrawPaintTransformer(visView.getPickedEdgeState()));

		// Default mouse support
		visView.setGraphMouse(graphMouse);

		// Finish layout algorithm if necessary
		if (layout instanceof KKLayout)
			((KKLayout) layout).lock(true);

		return visView;
	}

	/**
	 * 
	 * Creates an information (details) panel to show information about selected
	 * clusters.
	 * 
	 * @return A panel showing details of the selected cluster(s).
	 * 
	 */
	@SuppressWarnings("unchecked")
	private SelectionDetailsPane createRightPanel() {
		rightPane = new SelectionDetailsPane(globalContext);
		pickedState = visView.getPickedVertexState();

		pickedState.addItemListener(new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e) {
				rightPane.clear();
				if (pickedState.getPicked().size() == 1) {
					// One cluster selected, show cluster-specific info
					rightPane.showDetails(pickedState.getPicked().iterator().next(), globalClustering.getEventlog());
				} else if (pickedState.getPicked().size() > 1) {
					// Multiple clusters selected, show aggregate info
					rightPane.showDetails(pickedState.getPicked(), globalClustering.getEventlog());
				} else {
					// Nothing selected, hide panel
				}
				centerRight.setDividerLocation(centerRight.getMaximumDividerLocation() - 250);
			}
		});
		return rightPane;
	}
}