package org.processmining.traceclustering.visualization;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.Executor;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.ListSelectionModel;
import javax.swing.border.EmptyBorder;

import org.apache.commons.collections15.functors.ConstantTransformer;
import org.apache.commons.collections15.functors.MapTransformer;
import org.apache.commons.collections15.map.LazyMap;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.time.TimeSeriesCollection;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.traceclustering.algorithms.evolution.impl.SimilarityEvolution;
import org.processmining.traceclustering.connections.CaseClusteringConnection;
import org.processmining.traceclustering.models.impl.CaseCluster;
import org.processmining.traceclustering.models.impl.EvolutionaryCaseClustering;
import org.processmining.traceclustering.models.impl.TimeWindowType;
import org.processmining.traceclustering.models.perspectives.AbstractPerspective;
import org.processmining.traceclustering.parameters.CaseClusteringParameters;
import org.processmining.traceclustering.parameters.impl.TimeWindowParameters;
import org.processmining.traceclustering.plugins.EvolutionaryCaseClusteringPlugin;
import org.processmining.traceclustering.visualization.transformers.EdgeDrawPaintTransformer;
import org.processmining.traceclustering.visualization.transformers.VertexLabelTransformer;
import org.processmining.traceclustering.visualization.transformers.VertexShapeTransformer;
import org.processmining.traceclustering.visualization.transformers.VertexToolTipTransformer;

import com.fluxicon.slickerbox.factory.SlickerDecorator;
import com.fluxicon.slickerbox.factory.SlickerFactory;

import edu.uci.ics.jung.algorithms.layout.AggregateLayout;
import edu.uci.ics.jung.algorithms.layout.KKLayout;
import edu.uci.ics.jung.algorithms.layout.Layout;
import edu.uci.ics.jung.graph.util.Pair;
import edu.uci.ics.jung.visualization.DefaultVisualizationModel;
import edu.uci.ics.jung.visualization.VisualizationModel;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.DefaultModalGraphMouse;
import edu.uci.ics.jung.visualization.picking.PickedState;

/**
 * Visualizer for evolutionary trace clustering.
 * 
 * @author B.F.A. Hompes <b.f.a.hompes@tue.nl>
 *
 */
@Plugin(
		name = "Visualize Evolutionary Trace Clustering",
		returnLabels = { "Visualized Evolutionary Trace Clustering" },
		returnTypes = { JComponent.class },
		parameterLabels = { "Trace Clustering" },
		userAccessible = true)
@Visualizer
public class EvolutionaryCaseClusteringVisualization {

	UIPluginContext globalContext;
	EvolutionaryCaseClustering globalClustering;
	@SuppressWarnings("rawtypes")
	VisualizationViewer visViewLeft;
	@SuppressWarnings("rawtypes")
	VisualizationViewer visViewRight;
	ChartPanel chartPanel;
	DefaultModalGraphMouse<String, Object> graphMouse;
	SelectionDetailsPane rightPane;
	PickedState<CaseCluster> pickedStateLeft;
	PickedState<CaseCluster> pickedStateRight;

	// entireView holds dimension selection and buttons on left side, centerRight on right side
	JSplitPane entireView = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
	// centerRight holds center on left and detailsPane on right
	JSplitPane centerRight = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
	// center holds axis on bottom side, and two clusterings, one on the left and one on the right
	JPanel center = new JPanel(new GridBagLayout());

	private SlickerFactory slickerFactory;
	private SlickerDecorator slickerDecorator;

	@SuppressWarnings({ "unchecked", "rawtypes" })
	@PluginVariant(
			requiredParameterLabels = { 0 })
	public JPanel visualize(UIPluginContext context, EvolutionaryCaseClustering clustering) {
		/**************************************************************
		 * 
		 * Some global stuff
		 *
		 *************************************************************/

		globalContext = context;
		globalClustering = clustering;

		/**************************************************************
		 * 
		 * Default look & feel stuff & layout
		 *
		 *************************************************************/

		slickerFactory = SlickerFactory.instance();
		slickerDecorator = SlickerDecorator.instance();

		/**************************************************************
		 *
		 * Left panel: dimension selection and calculate button
		 *
		 **************************************************************/

		JPanel leftPane = new JPanel();
		leftPane.setPreferredSize(new Dimension(300, 80));
		leftPane.setBackground(Color.DARK_GRAY);
		leftPane.setForeground(Color.WHITE);
		leftPane.setLayout(new BoxLayout(leftPane, BoxLayout.PAGE_AXIS));
		leftPane.setBorder(new EmptyBorder(15, 5, 5, 15));

		// Mouse information
		JLabel mouseInfo = slickerFactory.createLabel("Select the mouse mode.");
		mouseInfo.setForeground(leftPane.getForeground());
		leftPane.add(mouseInfo);

		// Default mouse support
		graphMouse = new DefaultModalGraphMouse<String, Object>();
		JComboBox mouseMode = graphMouse.getModeComboBox();
		slickerDecorator.decorate(mouseMode);
		leftPane.add(mouseMode);

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));

		// Dimensions information
		JLabel dimensionInfo = slickerFactory.createLabel("Select the clustering dimensions.");
		dimensionInfo.setForeground(leftPane.getForeground());
		leftPane.add(dimensionInfo);

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));

		// Dimensions selection
		Collections.sort(globalClustering.getCaseClusteringParameters().getPerspectives(),
				new Comparator<AbstractPerspective>() {
					@Override
					public int compare(AbstractPerspective p1, AbstractPerspective p2) {
						return p1.getDescription().compareToIgnoreCase(p2.getDescription());
					}
				});
		JList dimensionsBox = new JList(globalClustering.getCaseClusteringParameters().getPerspectives().toArray());
		dimensionsBox.setCellRenderer(new PerspectiveCellRenderer());
		dimensionsBox.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		dimensionsBox.setBackground(leftPane.getBackground());
		dimensionsBox.setForeground(leftPane.getForeground());
		JScrollPane dimensionsScroller = new JScrollPane(dimensionsBox);
		dimensionsScroller.setBorder(null);
		slickerDecorator.decorate(dimensionsScroller, leftPane.getBackground(), Color.LIGHT_GRAY, Color.LIGHT_GRAY);
		leftPane.add(dimensionsScroller);

		// Time Window selection
		JLabel windowInfo = slickerFactory.createLabel("Window type and size.");
		windowInfo.setForeground(leftPane.getForeground());
		leftPane.add(windowInfo);
		final JComboBox windowType = new JComboBox();
		windowType.setModel(new DefaultComboBoxModel(TimeWindowType.values()));
		slickerDecorator.decorate(windowType);
		leftPane.add(windowType);
		final ProMTextField windowSize = new ProMTextField(1000 + "", "Window size (int).");
		leftPane.add(windowSize);
		final JButton calculateChange = slickerFactory.createButton("Calculate change in behavior over time");
		calculateChange.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				final TimeWindowParameters timeWindowParameters = new TimeWindowParameters();
				try {
					int size = 0;
					TimeWindowType type = (TimeWindowType) windowType.getSelectedItem();
					size = Integer.parseInt(windowSize.getText());
					timeWindowParameters.setType(type);
					timeWindowParameters.setWindowSize(size);
				} catch (Exception ex) {
					ex.printStackTrace();
				}

				final UIPluginContext child = globalContext.createChildContext("child");
				Executor exec = child.getExecutor();

				Runnable task = new Runnable() {
					@Override
					public void run() {
						// Change button name to show it's doing something
						calculateChange.setText("Calculating change over time...");

						TimeSeriesCollection timeWindowValues = SimilarityEvolution.calculate(
								globalClustering.getEventlog(), timeWindowParameters,
								globalClustering.getCaseClusteringParameters().getPerspectives());

						center.remove(chartPanel);
						center.invalidate();
						chartPanel = createChartPanel(timeWindowValues);
						GridBagConstraints c = new GridBagConstraints();
						c.gridx = 0;
						c.gridy = 1;
						c.gridwidth = 2;
						c.weightx = 1;
						c.weighty = 1;
						c.fill = GridBagConstraints.BOTH;
						c.anchor = GridBagConstraints.PAGE_END;
						center.add(chartPanel, c);
						chartPanel.setBorder(BorderFactory.createMatteBorder(2, 0, 0, 0, Color.DARK_GRAY));

						calculateChange.setText("Calculate change in behavior over time");
					}
				};

				exec.execute(task);
			}
		});

		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));
		leftPane.add(calculateChange);

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));

		// MCL Parameter inputs	
		CaseClusteringParameters parameters = globalClustering.getCaseClusteringParameters();
		JLabel expInfo = slickerFactory.createLabel("Expansion parameter.");
		expInfo.setForeground(leftPane.getForeground());
		leftPane.add(expInfo);
		final ProMTextField expansion = new ProMTextField(parameters.getMclParameters().getExpansionParameter() + "",
				"Expansion parameter (integer).");
		leftPane.add(expansion);

		JLabel inflInfo = slickerFactory.createLabel("Inflation parameter.");
		inflInfo.setForeground(leftPane.getForeground());
		leftPane.add(inflInfo);
		final ProMTextField inflation = new ProMTextField(parameters.getMclParameters().getInflationParameter() + "",
				"Inflation parameter (double).");
		leftPane.add(inflation);

		final JButton redoButtonLeft = slickerFactory.createButton("Calculate clustering");
		redoButtonLeft.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				final CaseClusteringParameters parameters = new CaseClusteringParameters(
						globalClustering.getCaseClusteringParameters());

				try {
					int expansionParameter = 0;
					double inflationParameter = 0;
					expansionParameter = Integer.parseInt(expansion.getText());
					inflationParameter = Double.parseDouble(inflation.getText());
					parameters.getMclParameters().setExpansionParameter(expansionParameter);
					parameters.getMclParameters().setInflationParameter(inflationParameter);
				} catch (Exception ex) {
					ex.printStackTrace();
				}

				final UIPluginContext child = globalContext.createChildContext("child");
				Executor exec = child.getExecutor();

				Runnable task = new Runnable() {
					@Override
					public void run() {
						// Change button name to show it's doing something
						redoButtonLeft.setText("Calculating clustering...");

						EvolutionaryCaseClusteringPlugin plugin = new EvolutionaryCaseClusteringPlugin();
						EvolutionaryCaseClustering clustering = plugin.doClusteringPublic(child,
								globalClustering.getEventlog(), parameters);

						// View the resulting clustering in the center pane of
						// the window
						center.remove(visViewLeft);
						center.invalidate();
						visViewLeft = createVisualisationViewer(clustering);
						GridBagConstraints c = new GridBagConstraints();
						c.gridx = 0;
						c.gridy = 0;
						c.weightx = 1;
						c.weighty = 2;
						c.fill = GridBagConstraints.BOTH;
						center.add(visViewLeft, c);
						visViewLeft.setBorder(BorderFactory.createMatteBorder(0, 0, 0, 2, Color.DARK_GRAY));
						center.revalidate();
						center.repaint();

						// Save this new clustering as a provided object in the resource view (not as favorite)
						child.getProvidedObjectManager().createProvidedObject("Clustering", clustering,
								EvolutionaryCaseClustering.class, globalContext);

						// Save connection between the log, the parameters and the resulting clustering
						child.getConnectionManager().addConnection(
								new CaseClusteringConnection(globalClustering.getEventlog(), clustering));

						centerRight.setRightComponent(createRightPanel());

						redoButtonLeft.setText("Calculate clustering");
					}
				};

				exec.execute(task);
			}
		});

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));
		leftPane.add(redoButtonLeft);

		final JButton redoButtonRight = slickerFactory.createButton("Calculate clustering to compare with");
		redoButtonRight.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				final CaseClusteringParameters parameters = new CaseClusteringParameters(
						globalClustering.getCaseClusteringParameters());
				try {
					int expansionParameter = 0;
					double inflationParameter = 0;
					expansionParameter = Integer.parseInt(expansion.getText());
					inflationParameter = Double.parseDouble(inflation.getText());
					parameters.getMclParameters().setExpansionParameter(expansionParameter);
					parameters.getMclParameters().setInflationParameter(inflationParameter);
				} catch (Exception ex) {
					ex.printStackTrace();
				}

				final UIPluginContext child = globalContext.createChildContext("child");
				Executor exec = child.getExecutor();

				Runnable task = new Runnable() {
					@Override
					public void run() {
						// Change button name to show it's doing something
						redoButtonRight.setText("Calculating clustering to compare with...");

						EvolutionaryCaseClusteringPlugin plugin = new EvolutionaryCaseClusteringPlugin();
						EvolutionaryCaseClustering clustering = plugin.doClusteringPublic(child,
								globalClustering.getEventlog(), parameters);

						// View the resulting clustering in the center pane of
						// the window
						center.remove(visViewRight);
						center.invalidate();
						visViewRight = createVisualisationViewer(clustering);
						GridBagConstraints c = new GridBagConstraints();
						c.gridx = 1;
						c.gridy = 0;
						c.weightx = 1;
						c.weighty = 2;
						c.fill = GridBagConstraints.BOTH;
						center.add(visViewRight, c);
						center.revalidate();
						center.repaint();

						// Save this new clustering as a provided object in the resource view (not as favorite)
						child.getProvidedObjectManager().createProvidedObject("Clustering", clustering,
								EvolutionaryCaseClustering.class, globalContext);

						// Save connection between the log, the parameters and the resulting clustering
						child.getConnectionManager().addConnection(
								new CaseClusteringConnection(globalClustering.getEventlog(), clustering));

						centerRight.setRightComponent(createRightPanel());

						redoButtonRight.setText("Calculate clustering to compare with");
					}
				};

				exec.execute(task);
			}
		});

		// Insert margin
		leftPane.add(Box.createRigidArea(new Dimension(0, 15)));
		leftPane.add(redoButtonRight);

		/**************************************************************
		 *
		 * Center panel: clustering visualization (graph) (initially empty)
		 *
		 **************************************************************/

		visViewLeft = createVisualisationViewer(globalClustering);
		visViewRight = createVisualisationViewer(globalClustering);
		chartPanel = new ChartPanel(null);

		/**************************************************************
		 *
		 * Right panel: details / selection
		 *
		 **************************************************************/

		rightPane = createRightPanel();

		/**************************************************************
		 *
		 * Wrap everything together and return
		 *
		 **************************************************************/

		centerRight.setLeftComponent(center);
		centerRight.setRightComponent(rightPane);

		entireView.setLeftComponent(leftPane);
		entireView.setRightComponent(centerRight);

		// Return graph in panel
		JPanel panel = new JPanel(new BorderLayout());
		entireView.setOpaque(false);
		centerRight.setOpaque(false);
		entireView.setOneTouchExpandable(true);
		centerRight.setOneTouchExpandable(true);
		panel.add(entireView, BorderLayout.CENTER);

		return panel;
	}

	/**
	 *
	 * Will create a graph to visualize the clustering using the JUNG library.
	 *
	 * @author B.F.A. Hompes
	 * @param clustering
	 * 
	 * @return A visualization viewer for the clustering
	 *
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private VisualizationViewer createVisualisationViewer(EvolutionaryCaseClustering clustering) {
		CaseClusterGraph graph = new CaseClusterGraph();
		Map vertexPaints = LazyMap.decorate(new HashMap(), new ConstantTransformer(new Color(240, 240, 240)));

		// Add vertices (clusters)
		for (CaseCluster cluster : clustering) {
			graph.addVertex(cluster);
		}

		// Add edges 
		// Similar traces are connected, only one edge is inserted for connections between clusters (traces in clusters)
		for (CaseCluster c1 : clustering) {
			cluster2: for (CaseCluster c2 : clustering) {
				if (!c1.equals(c2)) {
					// Only check one way since matrix is symmetrical over diagonal, hence matrix(s,t)=matrix(t,s)
					for (Integer source : c1) {
						for (Integer target : c2) {
							if (clustering.getSimilarityMatrix().get(source, target) > 1E-4) {
								graph.addEdge(new Edge(c1, c2, clustering.getSimilarityMatrix().get(source, target)),
										new Pair(c1, c2));
								break cluster2;
							}
						}
					}
				}
			}
		}

		// Create a general layout for our master graph, using the specified layout algorithm
		Layout layout = new KKLayout(graph);

		// Create an aggregate layout because we need subgraphs (for the clusters)
		AggregateLayout aglayout = new AggregateLayout(layout);

		// Visualization of the layout
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		//		Dimension preferredSize = new Dimension((int) (screenSize.width * 0.9), (int) (screenSize.height * 0.85));
		Dimension preferredSize = new Dimension((int) (screenSize.width * 0.9 / 2 - 250),
				(int) (screenSize.height * 0.85));
		VisualizationModel visualizationModel = new DefaultVisualizationModel(aglayout, preferredSize);
		VisualizationViewer visView = new VisualizationViewer(visualizationModel, preferredSize);

		// Tell the renderer to use our own customized color rendering
		visView.getRenderContext().setVertexFillPaintTransformer(MapTransformer.getInstance(vertexPaints));

		// Transformer for the vertices (label)
		visView.getRenderContext().setVertexLabelTransformer(new VertexLabelTransformer());

		// Transformer for the vertices (shape)
		visView.getRenderContext().setVertexShapeTransformer(new VertexShapeTransformer());

		// Transformer for the vertices (tooltip)
		visView.setVertexToolTipTransformer(new VertexToolTipTransformer());

		//  Transformer for the edges (color)
		visView.getRenderContext()
				.setEdgeDrawPaintTransformer(new EdgeDrawPaintTransformer(visView.getPickedEdgeState()));

		// Default mouse support
		visView.setGraphMouse(graphMouse);

		// Finish layout algorithm if necessary
		if (layout instanceof KKLayout)
			((KKLayout) layout).lock(true);

		return visView;
	}

	/**
	 * 
	 * Creates an information (details) panel to show information about selected
	 * clusters.
	 * 
	 * @return A panel showing details of the selected cluster(s).
	 * 
	 */
	@SuppressWarnings("unchecked")
	private SelectionDetailsPane createRightPanel() {
		rightPane = new SelectionDetailsPane(globalContext);
		pickedStateLeft = visViewLeft.getPickedVertexState();
		pickedStateRight = visViewRight.getPickedVertexState();

		pickedStateLeft.addItemListener(new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e) {
				if (pickedStateLeft.getPicked().size() == 1) {
					// One cluster selected, show cluster-specific info
					rightPane.showDetails(pickedStateLeft.getPicked().iterator().next(),
							globalClustering.getEventlog());
				} else if (pickedStateLeft.getPicked().size() > 1) {
					// Multiple clusters selected, show aggregate info
					rightPane.showDetails(pickedStateLeft.getPicked(), globalClustering.getEventlog());
				} else {
					// Nothing selected, hide panel
					rightPane.clear();
				}
				centerRight.setDividerLocation(centerRight.getMaximumDividerLocation() - 250);
				center.revalidate();
				center.repaint();
			}
		});
		pickedStateRight.addItemListener(new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e) {
				if (pickedStateRight.getPicked().size() == 1) {
					// One cluster selected, show cluster-specific info
					rightPane.showDetails(pickedStateRight.getPicked().iterator().next(),
							globalClustering.getEventlog());
				} else if (pickedStateRight.getPicked().size() > 1) {
					// Multiple clusters selected, show aggregate info
					rightPane.showDetails(pickedStateRight.getPicked(), globalClustering.getEventlog());
				} else {
					// Nothing selected, hide panel
					rightPane.clear();
				}
				centerRight.setDividerLocation(centerRight.getMaximumDividerLocation() - 250);
				center.revalidate();
				center.repaint();
			}
		});
		return rightPane;
	}

	private ChartPanel createChartPanel(TimeSeriesCollection timeWindowValues) {
		final DateAxis domainAxis = new DateAxis("Date");
		domainAxis.setDateFormatOverride(new SimpleDateFormat("MMM-yyyy", Locale.ENGLISH));
		final ValueAxis rangeAxis = new NumberAxis("Percentage");
		final TimeSeriesCollection data = timeWindowValues;
		final XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer();
		final XYPlot plot = new XYPlot(data, domainAxis, rangeAxis, renderer);
		JFreeChart chart = new JFreeChart("Change in Behavior", plot);
		chart.getLegend().setVisible(false);
		JFreeChartCustomThemes.getPrintTheme().apply(chart);
		renderer.setSeriesShapesVisible(0, true);
		//		renderer.setSeriesToolTipGenerator(0, new XYToolTipGenerator() {
		//			public String generateToolTip(XYDataset dataset, int series, int item) {
		//				Number x1 = dataset.getX(series, item);
		//				Number y1 = dataset.getY(series, item);
		//				StringBuilder stringBuilder = new StringBuilder();
		//				stringBuilder.append(String.format("<html><p style='color:#0000ff;'>Serie: '%s'</p>",
		//						dataset.getSeriesKey(series)));
		//				stringBuilder.append(String.format("Date:'%d'<	br/>", x1.intValue()));
		//				stringBuilder.append(String.format("Change:'%d'", y1.intValue()));
		//				stringBuilder.append("</html>");
		//				return stringBuilder.toString();
		//			}
		//		});
		rangeAxis.setTickMarksVisible(true);
		final ChartPanel chartPanel = new ChartPanel(chart);
		chartPanel.setMouseZoomable(true, false);
		return chartPanel;
	}
}