package org.processmining.traceclustering.visualization;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashSet;
import java.util.Set;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.border.EmptyBorder;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.traceclustering.models.impl.CaseCluster;
import org.processmining.traceclustering.utils.LogUtils;

import com.fluxicon.slickerbox.factory.SlickerDecorator;
import com.fluxicon.slickerbox.factory.SlickerFactory;

/**
 * Class for pane that shows details (annotations) of a selection from a
 * clustering.
 * 
 * @author B.F.A. Hompes <b.f.a.hompes@tue.nl>
 *
 */
public class SelectionDetailsPane extends JPanel {

	private static final long serialVersionUID = 9043788718544268406L;

	private UIPluginContext context;
	private SlickerFactory slickerFactory;
	private SlickerDecorator slickerDecorator;

	public SelectionDetailsPane(UIPluginContext context) {
		this.context = context;
		setPreferredSize(new Dimension(250, 80));
		setBackground(Color.DARK_GRAY);
		setForeground(Color.WHITE);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		slickerFactory = SlickerFactory.instance();
		slickerDecorator = SlickerDecorator.instance();

		setBorder(new EmptyBorder(15, 5, 5, 15));

		clear();
	}

	public void clear() {
		setEnabled(false);
		setVisible(false);
	}

	public void showDetails(final CaseCluster cluster, final XLog eventlog) {
		HashSet<CaseCluster> clusters = new HashSet<CaseCluster>();
		clusters.add(cluster);
		showDetails(clusters, eventlog);
	}

	public void showDetails(final Set<CaseCluster> clusters, final XLog eventlog) {
		this.removeAll();

		DefaultListModel<String> listModel = new DefaultListModel<String>();

		JList<String> annotations = new JList<String>(listModel);
		annotations.setLayoutOrientation(JList.VERTICAL);
		//		annotations.setVisibleRowCount(-1);
		annotations.setBackground(Color.DARK_GRAY);
		annotations.setForeground(Color.WHITE);

		JScrollPane listScroller = new JScrollPane(annotations);
		listScroller.setBorder(null);
		listScroller.setPreferredSize(new Dimension(250, 80));

		JButton exportLogButton;

		if (clusters.size() == 1) {
			// We have selected one cluster
			final CaseCluster cluster = clusters.iterator().next();

			// Show all kinds of information of the selected cluster.
			JLabel info = slickerFactory.createLabel("Cluster information");
			info.setForeground(this.getForeground());
			add(info);

			// Add margin
			add(Box.createRigidArea(new Dimension(0, 15)));

			// Add size
			listModel.addElement("Size: " + cluster.size());

			// Add annotations
			for (org.processmining.traceclustering.models.perspectives.AbstractPerspective dimension : cluster
					.getAnnotation().keySet()) {
				listModel.addElement(dimension.getDescription() + ": " + cluster.getAnnotation().get(dimension));
			}

			// Button to export selected cluster
			exportLogButton = slickerFactory.createButton("Export filtered log to ProM workspace");
			exportLogButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					XLog filteredLog = LogUtils.filterLog(eventlog, cluster);
					context.getProvidedObjectManager().createProvidedObject(
							"Filtered log from " + XConceptExtension.instance().extractName(filteredLog) + " ("
									+ filteredLog.size() + " cases)",
							filteredLog, XLog.class, context);
					context.getGlobalContext().getResourceManager().getResourceForInstance(filteredLog)
							.setFavorite(true);
				}
			});
		} else {
			// We have selected multiple clusters

			// Show all kinds of information of the selected cluster.
			JLabel info = slickerFactory.createLabel("Cluster information");
			info.setForeground(this.getForeground());
			add(info);

			// Add margin
			add(Box.createRigidArea(new Dimension(0, 15)));

			// Add size
			listModel.addElement("You have selected " + clusters.size() + " clusters.");
			int size = 0;
			for (CaseCluster cluster : clusters) {
				size += cluster.size();
			}
			listModel.addElement("Total size: " + size);
			listModel.addElement("");

			// Add annotations
			final CaseCluster joinedCluster = new CaseCluster();
			for (CaseCluster cluster : clusters) {
				joinedCluster.addAll(cluster);
				for (org.processmining.traceclustering.models.perspectives.AbstractPerspective dimension : cluster
						.getAnnotation().keySet()) {
					listModel.addElement(dimension.getDescription() + ": " + cluster.getAnnotation().get(dimension));
				}
				// Add separator (empty line) between clusters
				listModel.addElement("-");
			}

			// Button to export selected cluster
			exportLogButton = slickerFactory.createButton("Export filtered log to ProM workspace");
			exportLogButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					XLog filteredLog = LogUtils.filterLog(eventlog, joinedCluster);
					context.getProvidedObjectManager().createProvidedObject(
							"Filtered log from " + XConceptExtension.instance().extractName(filteredLog) + " ("
									+ joinedCluster.size() + " cases)",
							filteredLog, XLog.class, context);
					context.getGlobalContext().getResourceManager().getResourceForInstance(filteredLog)
							.setFavorite(true);
				}
			});
		}

		// Add description
		slickerDecorator.decorate(listScroller, getBackground(), Color.LIGHT_GRAY, Color.LIGHT_GRAY);
		add(listScroller);

		// Add margin
		add(Box.createRigidArea(new Dimension(0, 15)));

		// Add button
		add(exportLogButton);

		// Add margin
		add(Box.createRigidArea(new Dimension(0, 15)));

		// Show pane
		setEnabled(true);
		setVisible(true);
	}

}